
from fnmatch import filter as flt  # Native filter() function is used as well
from typing import Type, Union, List, Tuple
from dataclasses import dataclass
import pkg_resources
import sys
import os
import re

from ase import Atom, Atoms
import pandas as pd
import numpy as np

# Pull in molar mass data from separate csv file
mm_data = pkg_resources.resource_stream(__name__, 'data/mm_of_elements.csv')
molarmass_df = pd.read_csv(mm_data, encoding=sys.stdout.encoding, index_col=0)
# TODO - see if we can pull this in relatively with just pandas - I don't remember


@dataclass
class Control:
    calculation: str = 'relax'
    prefix: str = 'untitled'
    outdir: str = None
    pseudo_dir: str = None
    nstep: int = None
    verbosity: str = None
    restart_mode: str = None
    dt: float = None
    max_seconds: float = None
    forc_conv_thr: float = None
    etot_conv_thr: float = None
    lelfield: Union[bool, str] = None
    dipfield: Union[bool, str] = None
    tefield: Union[bool, str] = None


@dataclass
class System:
    ibrav: int = 0
    celldm: List[float] = None
    A: float = None
    B: float = None
    C: float = None
    cosAB: float = None
    cosAC: float = None
    cosBC: float = None
    nat: int = None
    ntyp: int = None
    nbnd: int = None
    tot_charge: float = None
    ecutwfc: float = None
    ecutrho: float = None
    nosym: Union[bool, str] = None
    nosym_evc: Union[bool, str] = None
    occupations: str = 'smearing'
    degauss: float = 0.02
    smearing: str = 'mv'
    nspin: int = None
    input_dft: str = None
    assume_isolated: str = None
    esm_bc: str = None
    starting_magnetizations: List[int] = None
    noncolin: Union[bool, str] = None
    edir: int = None
    emaxpos: float = None
    eopreg: float = None
    eamp: float = None


@dataclass
class Electrons:
    electron_maxstep: int = None
    scf_must_converge: Union[bool, str] = None
    conv_thr: float = None
    adaptive_thr: Union[bool, str] = None
    conv_thr_init: float = None
    mixing_mode: str = None
    mixing_beta: float = 0.7
    mixing_ndim: int = None
    diagonalization: str = None


@dataclass
class Cell:
    cell_dynamics: str = None
    press: float = None
    wmass: float = None
    cell_factor: float = None
    cell_dofree: str = None


@dataclass
class Job:
    job_type: str = 'slurm'
    nodes: int = 1
    ntasks: int = 16
    partition: str = 'general'
    time: int = 50
    memory: int = 100
    mail_type: List[str] = None  # Options: ( NONE, BEGIN, END, FAIL, REQUEUE, ALL )
    mail_user: str = None
    exec: str = 'pw.x'
    nk: int = None
    input: str = None
    output: str = None


def position(geometry: Union[Type[Atom], Type[Atoms]] = None) -> Tuple[int, int, str]:

    atomic_positions = ''
    positions = geometry.get_positions().tolist()
    symbols = geometry.get_chemical_symbols()
    unique_symbols = list(set(symbols))
    atom_count = len(positions)

    for atom_set in zip(symbols, positions):
        atomic_positions += f'   {atom_set[0]}\t{np.round(atom_set[1][0], 8):.8f}'
        atomic_positions += f'\t{np.round(atom_set[1][1], 8):.8f}\t{np.round(atom_set[1][2], 8):.8f}\n'

    return atom_count, len(unique_symbols), atomic_positions
    # nat, ntyp, atomic_positions


def species(geometry: Union[Type[Atom], Type[Atoms]] = None, pseudo_dir: str = None) -> List[List[str]]:

    symbols = geometry.get_chemical_symbols()
    unique_symbols = sorted(list(set(symbols)))

    list_upf = flt(os.listdir(pseudo_dir), '*.[Uu][Pp][Ff]')
    species_list = []

    for symbol in unique_symbols:

        r = re.compile(rf'{symbol}[_|.|-]\S+\Z', flags=re.IGNORECASE)
        match = list(filter(r.match, list_upf))[0]
        mw_species = molarmass_df.loc[symbol][0]

        species_list.append([symbol, mw_species, match])

    return species_list


def cell_parameters(geometry: Union[Type[Atom], Type[Atoms]] = None) -> str:

    supercell = geometry.get_cell()
    cell_string = ''

    for dimension in supercell:
        cell_string += f'{dimension[0]:.14f}\t{dimension[1]:.14f}\t{dimension[2]:.14f}\n'

    return cell_string


def lock_atoms(lock: Union[str, Tuple[int]] = None, which: Tuple[int] = (0, 0, 0), positions: str = None) -> str:

    position_index = list(range(len(positions.splitlines())))

    if 'first' in lock:
        lock = lock.lstrip('first')
        lock = tuple(position_index[:int(lock) + 1])
    elif 'last' in lock:
        lock = lock.lstrip('last')
        lock = tuple(position_index[-int(lock):])

    new_positions = []
    for index, line in enumerate(positions.splitlines()):
        if index in lock:
            line = re.sub(r' \d \d \d', '', line)
            for value in which:
                line += f' {value}'

        new_positions.append(line)

    new_positions = '\n'.join(new_positions)

    return new_positions



"""
Alex Summers
The current function of this script is to create a Relax object to handle relevant information
and create a generic input file and bash script for use with Quantum Espresso.
For more information see the GitHub repository at https://github.com/ajsummers/ninia.
"""

from ninia.utils import Control, System, Electrons, Cell, Job
from typing import Type, Union, List, Tuple
from jinja2 import Environment, BaseLoader
from ase import Atom, Atoms
from ninia import utils
import pkg_resources
import os

starting_dir = os.getcwd()

input_string = pkg_resources.resource_string(__name__, 'input/relax.jinja2')
input_template = Environment(loader=BaseLoader).from_string(input_string.decode('utf-8'))
slurm_string = pkg_resources.resource_string(__name__, 'input/slurm.jinja2')
slurm_template = Environment(loader=BaseLoader).from_string(slurm_string.decode('utf-8'))


class Relax:

    def __init__(self, control: Type[Control] = Control(), system: Type[System] = System(),
                 electrons: Type[Electrons] = Electrons(), cell: Type[Cell] = Cell(), job: Type[Job] = Job(),
                 geometry: Union[Type[Atom], Type[Atoms]] = None, input_dir: str = None,
                 k_points: Tuple[int] = (1, 1, 1, 0, 0, 0)):

        # Initialize class parameters

        self.control = control
        self.system = system
        self.electrons = electrons
        self.cell = cell
        self.job = job
        self.geometry = geometry
        self.input_dir = input_dir
        self.k_points = k_points

        self.cell_parameters = None
        self.atomic_species = None
        self.atomic_positions = None

    def set_atomic_info(self, geometry: Union[Type[Atom], Type[Atoms]] = None) -> None:

        if (self.geometry is None) and (geometry is None):
            raise RuntimeError('Need to define geomtry info (ase.Atoms object).')
        elif geometry is not None:
            self.geometry = geometry

        self.set_pseudodir()

        self.system.nat, self.system.ntyp, self.atomic_positions = utils.position(self.geometry)
        self.atomic_species = utils.species(self.geometry, pseudo_dir=self.control.pseudo_dir)
        self.cell_parameters = utils.cell_parameters(self.geometry)

    def set_pseudodir(self, pseudo_dir: str = None):

        if (self.control.pseudo_dir is None) and (pseudo_dir is None):
            raise RuntimeError('Need to define pseudo_dir.')
        elif pseudo_dir is not None:
            self.control.pseudo_dir = pseudo_dir

        if not os.path.isabs(self.control.pseudo_dir):
            self.control.pseudo_dir = os.path.join(starting_dir, self.control.pseudo_dir)

        self.check_directory_(self.control.pseudo_dir)

    def set_outdir(self, outdir: str = None):
        if (self.control.outdir is None) and (outdir is None):
            raise RuntimeError('Need to define outdir.')
        elif outdir is not None:
            self.control.outdir = outdir

        if not os.path.isabs(self.control.outdir):
            self.control.outdir = os.path.join(starting_dir, self.control.outdir)

        self.check_directory_(self.control.outdir)

    def set_inputdir(self, input_dir: str = None):
        if (self.input_dir is None) and (input_dir is None):
            self.input_dir = starting_dir
        elif input_dir is not None:
            self.input_dir = input_dir

        if not os.path.isabs(self.input_dir):
            self.input_dir = os.path.join(starting_dir, self.input_dir)

        self.check_directory_(self.input_dir)

    @staticmethod
    def check_directory_(directory):

        if not os.path.isdir(directory):
            raise NotADirectoryError(f'{directory} is not a valid directory.')

    def set_prefix_(self):

        input_file = os.path.join(self.input_dir, f'{self.control.prefix}.i')
        if not (os.path.isfile(input_file)):
            save_file = os.path.join(self.control.outdir, f'{self.control.prefix}.save')
            wfc_file = os.path.join(self.control.outdir, f'{self.control.prefix}.wfc1')
            if os.path.isdir(save_file) or os.path.isfile(wfc_file):
                self.control.prefix += '_1'

            i = 1
            save_file = os.path.join(self.control.outdir, f'{self.control.prefix}.save')
            wfc_file = os.path.join(self.control.outdir, f'{self.control.prefix}.wfc1')
            while os.path.isdir(save_file) or os.path.isfile(wfc_file):
                i += 1
                self.control.prefix = f'{self.control.prefix.rstrip("0123456789")}{i}'
                save_file = os.path.join(self.control.outdir, f'{self.control.prefix}.save')
                wfc_file = os.path.join(self.control.outdir, f'{self.control.prefix}.wfc1')

            return True

        else:

            return False

    def create_input(self, control: Type[Control] = None, system: Type[System] = None,
                     electrons: Type[Electrons] = None, cell: Type[Cell] = None,
                     geometry: Union[Type[Atom], Type[Atoms]] = None):

        if control is not None:
            self.control = control
        if system is not None:
            self.system = system
        if electrons is not None:
            self.electrons = electrons
        if cell is not None:
            self.cell = cell
        if geometry is not None:
            self.geometry = geometry
            self.set_atomic_info(geometry)

        if self.atomic_positions is None:
            self.set_atomic_info(self.geometry)

        self.set_outdir()
        self.set_inputdir()

        if self.set_prefix_():

            input_content = input_template.render(control=self.control, system=self.system, electrons=self.electrons,
                                                  cell=self.cell, atomic_species=self.atomic_species,
                                                  cell_parameters=self.cell_parameters,
                                                  atomic_positions=self.atomic_positions,
                                                  k_points=self.k_points)

            input_file = os.path.join(self.input_dir, f'{self.control.prefix}.i')
            with open(input_file, 'w') as f:
                f.write(input_content)

            print(f'Created input file {input_file}')

        else:

            input_file = os.path.join(self.input_dir, f'{self.control.prefix}.i')
            print(f'Muted redundant input file {input_file}')

    def create_job(self, job: Type[Job] = None):

        if job is not None:
            self.job = job

        self.check_directory_(self.input_dir)
        if self.job.input is None:
            self.job.input = os.path.join(self.input_dir, f'{self.control.prefix}.i')
        if self.job.output is None:
            self.job.output = os.path.join(self.input_dir, f'{self.control.prefix}.out')

        job_file = os.path.join(self.input_dir, f'{self.control.prefix}.sh')
        if not (os.path.isfile(job_file)):
            job_content = slurm_template.render(control=self.control, job=self.job)

            with open(job_file, 'w') as f:
                f.write(job_content)

            os.chdir(starting_dir)
            print(f'Created bash file {job_file}')

        else:

            print(f'Muted redundant bash file {job_file}')

    def lock_atoms(self, lock: Union[str, Tuple[int]] = None, which: Tuple[int] = (0, 0, 0)) -> None:

        if self.atomic_positions is None:
            self.set_atomic_info(self.geometry)

        self.atomic_positions = utils.lock_atoms(lock, which, self.atomic_positions)




# Copyright 2022 Alex Summers
# See LICENSE file for more information

import sys

if sys.version_info[0] == 2:
    raise ImportError('Ninia requires Python 3.7. This is Python 2.')

__all__ = ['Relax']
__version__ = '0.0.78'

from ninia.relax import Relax



"""
Alex Summers
The current function of this script is to create a Relax object to handle relevant information
and create a generic input file and bash script for use with Quantum Espresso.
For more information see the GitHub repository at https://github.com/ajsummers/ninia.
"""


import pkg_resources
import warnings
import sys
import os
import re

from fnmatch import filter as flt  # Native filter() function is used as well
import pandas as pd
import numpy as np


starting_dir = os.getcwd()

# Pull in molar mass data from separate csv file
mm_data = pkg_resources.resource_stream(__name__, 'data/mm_of_elements.csv')
molarmass_df = pd.read_csv(mm_data, encoding=sys.stdout.encoding, index_col=0)


class Relax:

    def __init__(self, prefix=None, functional=None):

        # TODO - class parameters prefix, functional, and pseudo_dir not getting set if None type

        if prefix is None:
            warnings.warn('Prefix not defined. Setting to "untitled".', UserWarning)
        self.prefix = prefix

        if functional is None:
            warnings.warn('Functional not defined. Using BEEF-vdW by default.', UserWarning)
            self.functional = 'beef'
        else:
            self.functional = functional

        # TODO - Add additional checks for pseudopotential directory?
        # Initialize class parameters

        self.pseudo_dir = None

        self.geometry = None
        self.output_dir = None
        self.pseudo_dir = None
        self.input_dir = None
        self.nstep = None

        self.num_atoms = None
        self.num_elem = None
        self.ecutwfc = None
        self.ecutrho = None

        self.conv_thr = None
        self.mixing_beta = None
        self.electron_maxstep = None

        self.atomic_species = None
        self.cell_parameters = None
        self.atomic_positions = None
        self.k_points = None

        self.job_type = None
        self.partition = None

        self.memory = None
        self.cpus = None
        self.hours = None
        self.nk = 1

        self.locked = False
        self.catkit = True

    def get_position_info_(self, ase_object):

        atomic_positions = ''
        positions = ase_object.get_positions().tolist()
        symbols = ase_object.get_chemical_symbols()
        unique_symbols = list(set(symbols))
        atom_count = len(positions)

        if self.locked is False:
            locked = np.zeros((atom_count, 3))
        elif len(self.locked) != atom_count:
            raise ValueError('Locked dimensions do not match position dimensions.')
        else:
            locked = self.locked

        for atom_set in zip(symbols, positions, locked):
            atomic_positions += f'   {atom_set[0]}\t{np.round(atom_set[1][0], 8):.8f}'
            atomic_positions += f'\t{np.round(atom_set[1][1], 8):.8f}\t{np.round(atom_set[1][2], 8):.8f}'
            atomic_positions += f' {atom_set[2][0]:1.0f} {atom_set[2][1]:1.0f} {atom_set[2][2]:1.0f}\n'

        self.num_atoms = atom_count
        self.num_elem = len(unique_symbols)
        self.atomic_positions = atomic_positions

        return unique_symbols

    def get_species_info_(self, species_list):

        os.chdir(self.pseudo_dir)
        list_upf = flt(os.listdir('.'), '*.[Uu][Pp][Ff]')
        species_string = ''

        for species in species_list:

            r = re.compile(rf'{species}[_|.]\S+\Z', flags=re.IGNORECASE)
            match = list(filter(r.match, list_upf))[0]
            mw_species = molarmass_df.loc[species][0]

            species_string += f'   {species}\t{mw_species}\t{match}\n'

        os.chdir(starting_dir)
        self.atomic_species = species_string

    def get_cell_parameters_(self, ase_object):

        supercell = ase_object.get_cell()
        if not self.catkit:
            supercell[2][2] = 2 * np.max(ase_object.get_positions().T[2])
        cell_parameters = ''

        for dimension in supercell:
            cell_parameters += f'{dimension[0]:.14f}\t{dimension[1]:.14f}\t{dimension[2]:.14f}\n'

        self.cell_parameters = cell_parameters

    @staticmethod
    def check_directory_(directory):

        if not os.path.isdir(directory):
            raise NotADirectoryError(f'{directory} is not a valid directory.')

    def set_prefix_(self):

        # TODO - test if this actually omits repeat input files

        os.chdir(self.input_dir)
        if not (os.path.isfile(f'{self.prefix}.i')):
            os.chdir(self.output_dir)
            if (os.path.isdir(f'{self.prefix}.save')) or (os.path.isfile(f'{self.prefix}.wfc1')):
                self.prefix += '_1'

            i = 1
            while (os.path.isdir(f'{self.prefix}.save')) or (os.path.isfile(f'{self.prefix}.wfc1')):
                i += 1
                self.prefix = f'{self.prefix.rstrip("0123456789")}{i}'

            os.chdir(starting_dir)
            return True

        else:
            os.chdir(starting_dir)
            return False

    def set_directories(self, inputdir=None, outputdir=None, pseudodir=None):

        if (self.pseudo_dir is None) and (pseudodir is None):
            warnings.warn('Pseudopotential directory still not specified. Loading '
                          'geometry will likely result in an error.', UserWarning)
        elif (self.pseudo_dir is None) and (pseudodir is not None):
            self.check_directory_(pseudodir)
            self.pseudo_dir = pseudodir
        elif (self.pseudo_dir is not None) and (pseudodir is not None):
            print(f'Changing pseudodirectory to {pseudodir}.')
            self.check_directory_(pseudodir)
            self.pseudo_dir = pseudodir

        if inputdir is not None:
            self.check_directory_(inputdir)
            self.input_dir = inputdir
        else:
            self.input_dir = starting_dir

        if outputdir is not None:
            self.check_directory_(outputdir)
            self.output_dir = outputdir

    def load_geometry(self, ase_object):

        self.geometry = ase_object
        species_list = self.get_position_info_(self.geometry)
        self.get_species_info_(species_list)
        self.get_cell_parameters_(self.geometry)

    def set_parameters(self, ecutwfc=None, ecutrho=None, conv_thr=None, mixing_beta=None, k_points=None,
                       electron_maxstep=None, functional=None, nstep=None, locked=None, catkit=False):

        if (self.functional is None) and (functional is None):
            warnings.warn('Functional is still not specified. Creating input will likely result in an error.',
                          UserWarning)
        elif (self.functional is None) and (functional is not None):
            self.functional = functional
        elif (self.functional is not None) and (functional is not None):
            print(f'Changing functional to {functional}')

        if (self.ecutwfc is None) and (ecutwfc is None):
            self.ecutwfc = 30.0
        elif ecutwfc is not None:
            self.ecutwfc = ecutwfc

        if (self.ecutrho is None) and (ecutrho is None):
            self.ecutrho = 4 * self.ecutwfc
        elif ecutrho is not None:
            self.ecutrho = ecutrho

        if (self.conv_thr is None) and (conv_thr is None):
            self.conv_thr = '1.0d-8'
        elif conv_thr is not None:
            self.conv_thr = str(conv_thr).replace('e', 'd')

        if (self.mixing_beta is None) and (mixing_beta is None):
            self.mixing_beta = '0.7d0'
        elif mixing_beta is not None:
            self.mixing_beta = str(mixing_beta).replace('e', 'd')

        if (self.k_points is None) and (k_points is None):
            self.k_points = ' 4 4 4 0 0 0'
        elif k_points is not None:
            k_string = ''
            for point in k_points:
                k_string += f' {point}'
            self.k_points = k_string

        if (self.electron_maxstep is None) and (electron_maxstep is None):
            self.electron_maxstep = 100
        elif electron_maxstep is not None:
            self.electron_maxstep = electron_maxstep

        if (self.nstep is None) and (nstep is None):
            self.nstep = 100
        elif nstep is not None:
            self.nstep = nstep

        if (self.locked is None) and (locked is None):
            self.locked = False
        elif locked is not None:
            self.locked = locked

        self.catkit = catkit

    def create_input(self):  # Note - removed 'repeat' parameter. What was this for?

        if self.set_prefix_():

            runtime_error = ''
            runtime_error += ' self.geometry' if not self.geometry else ''
            runtime_error += ' self.prefix' if not self.prefix else ''
            runtime_error += ' self.output_dir' if not self.output_dir else ''
            runtime_error += ' self.pseudo_dir' if not self.pseudo_dir else ''
            runtime_error += ' self.input_dir' if not self.input_dir else ''
            runtime_error += ' self.num_atoms' if not self.num_atoms else ''
            runtime_error += ' self.num_elem' if not self.num_elem else ''
            runtime_error += ' self.ecutwfc' if not self.ecutwfc else ''
            runtime_error += ' self.functional' if not self.functional else ''
            runtime_error += ' self.conv_thr' if not self.conv_thr else ''
            runtime_error += ' self.mixing_beta' if not self.mixing_beta else ''
            runtime_error += ' self.atomic_species' if not self.atomic_species else ''
            runtime_error += ' self.cell_parameters' if not self.cell_parameters else ''
            runtime_error += ' self.atomic_positions' if not self.atomic_positions else ''
            runtime_error += ' self.k_points' if not self.k_points else ''

            if runtime_error is not '':
                raise RuntimeError(f'Missing{runtime_error}')

            input_template = pkg_resources.resource_string(__name__, 'input/relax.i').decode(sys.stdout.encoding)
            compiled_fstring = compile(input_template, '<fstring_from_file', 'eval')
            formatted_relax = eval(compiled_fstring)

            os.chdir(self.input_dir)
            with open(f'{self.prefix}.i', 'w+') as f:
                f.write(formatted_relax)

            os.chdir(starting_dir)
            print(f'Created input file {self.prefix}.i')

        else:

            print(f'Muted redundant input file {self.prefix}.i')

    def create_job(self, job_type=None, partition=None, memory=None, cpus=None, hours=None):

        if job_type is None:
            self.job_type = 'pbs.sh'
        else:
            self.job_type = job_type

        if partition is None:
            self.partition = 'general'
        else:
            self.partition = partition

        if memory is None:
            self.memory = 50
        else:
            self.memory = memory

        if cpus is None:
            self.cpus = 8
        else:
            self.cpus = cpus

        if hours is None:
            self.hours = 100
        else:
            self.hours = hours

        bash_template = pkg_resources.resource_string(__name__, f'input/{self.job_type}').decode(sys.stdout.encoding)
        compiled_bash_fstring = compile(bash_template, '<fstring_from_file', 'eval')
        formatted_bash_relax = eval(compiled_bash_fstring)

        os.chdir(self.input_dir)
        if not (os.path.isfile(f'{self.prefix}.sh')):

            with open(f'{self.prefix}.sh', 'w+') as f:
                f.write(formatted_bash_relax)

            os.chdir(starting_dir)
            print(f'Created bash file {self.prefix}.sh')

        else:

            print(f'Muted redundant bash file {self.prefix}.sh')
